/*
 * Copyright (c) 1999 by Sun Microsystems, Inc.
 * All rights reserved.
 */

/* 
 * Portions may be based on code from the linux ext2 filesystem 
 * 
 *  Copyright (C) 1993  Stephen Tweedie (sct@dcs.ed.ac.uk)
 *  Copyright (C) 1992  Remy Card (card@masi.ibp.fr)
 *                      Laboratoire MASI - Institut Blaise Pascal
 *                      Universite Pierre et Marie Curie (Paris VI)
 * Copyright (C) 1991, 1992  Linus Torvalds
 */

#pragma ident	"%Z%%M%	%I%	%E% SMI"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kmem.h>
#include <sys/user.h>
#include <sys/proc.h>
#include <sys/cred.h>
#include <sys/disp.h>
#include <sys/buf.h>
#include <sys/vfs.h>
#include <sys/vnode.h>
#include <sys/fdio.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/conf.h>
#undef NFSCLIENT
#include <sys/statvfs.h>
#include <sys/mount.h>
#include <sys/pathname.h>
#include <sys/cmn_err.h>
#include <sys/debug.h>
#include <sys/sysmacros.h>
#include <sys/conf.h>
#include <sys/mkdev.h>
#include <sys/swap.h>
#include <sys/sunddi.h>
#include <sys/dktp/fdisk.h>
#include <fs/fs_subr.h>
#include <sys/modctl.h>
#include <sys/vol.h>
#include <sys/dkio.h>
#include <sys/open.h>
#include <sys/mntent.h>
#include <ext2fs.h>


static int ext2fs_init(struct vfssw *, int);
static int ext2fs_mount(struct vfs *, struct vnode *, struct mounta *,
    struct cred *);
static int
ext2fs_unmount(struct vfs *vfsp,
#ifdef MS_FORCE
    int flag,
#endif
    struct cred *cr);
static int ext2fs_root(struct vfs *, struct vnode **);
static int ext2fs_statvfs(struct vfs *, struct statvfs64 *);
static int ext2fs_sync(struct vfs *, short, struct cred *);
static int ext2fs_vget(struct vfs *vfsp, struct vnode **vpp, struct fid *fidp);
static void ihfini(void);

#ifndef MS_FORCE
#define vfs_devismounted(DEV) (vfs_devsearch(DEV) != NULL)
#endif

static int ext2fstype;
static kmutex_t ext2fs_lock;
static struct ext2fs *ext2fs_mounttab;

#ifdef VSW_HASPROTO
/*
 * ext2fs mount options table
 */

static char *rw_cancel[] = {MNTOPT_RO, NULL};
static char *ro_cancel[] = {MNTOPT_RW, NULL};

static mntopt_t mntopts[] = {
/*
 *	option name		cancel option	default arg	flags
 *		opt data
 */
	{ MNTOPT_RW,		rw_cancel,	NULL,		MO_DEFAULT,
		NULL },
	{ MNTOPT_RO,		ro_cancel,	NULL,		0,
		NULL }
};

static mntopts_t ext2fs_mntopts = {
	sizeof (mntopts) / sizeof (mntopt_t),
	mntopts
};
#endif

struct vfsops ext2fs_vfsops = {
	ext2fs_mount,
	ext2fs_unmount,
	ext2fs_root,
	ext2fs_statvfs,
	ext2fs_sync,
	ext2fs_vget,
	fs_nosys,       /* vfs_mountroot */
	fs_nosys,        /* vfs_swapvp */
#ifdef MS_FORCE
	fs_freevfs
#endif
};

static struct vfssw vfw = {
	"ext2fs",
	ext2fs_init,
	&ext2fs_vfsops,
#ifdef VSW_HASPROTO
	VSW_HASPROTO,
	&ext2fs_mntopts
#endif
};

static struct modlfs modlfs = {
	&mod_fsops,
	"Linux Second Extended Filesystem",
	&vfw
};

static struct modlinkage modlinkage = {
	MODREV_1,
	&modlfs,
	NULL
};

int
_init(void)
{
	int	error;

	mutex_init(&ext2fs_lock, NULL, MUTEX_DEFAULT, NULL);
	error = mod_install(&modlinkage);
	if (error) {
		mutex_destroy(&ext2fs_lock);
	}
	cmn_err(CE_NOTE,
	    "This version of ext2fs is not supported, and will likely need "
	    "recompilation after Upgrades and Patches."
	    );
	return (error);
}

int
_fini(void)
{
	int	error;

	error = mod_remove(&modlinkage);
	if (error)
		return (error);
	ihfini();
	mutex_destroy(&ext2fs_lock);
	return (0);
}

int
_info(struct modinfo *modinfop)
{
	return (mod_info(&modlinkage, modinfop));
}

/* ARGSUSED */
static int
ext2fs_inode_cache_constructor(void *buf, void *cdrarg, int kmflags)
{
	struct inode *ip = buf;
	struct vnode *vp;

	vp = ITOV(ip);

	mutex_init(&vp->v_lock, NULL, MUTEX_DEFAULT, NULL);
	mutex_init(&ip->i_tlock, NULL, MUTEX_DEFAULT, NULL);
	rw_init(&ip->i_rwlock, NULL, RW_DEFAULT, NULL);

	cv_init(&vp->v_cv, NULL, CV_DEFAULT, NULL);

	return (0);
}

/* ARGSUSED */
static void
ext2fs_inode_cache_destructor(void *buf, void *cdrarg)
{
	struct inode *ip = buf;
	struct vnode *vp;

	vp = ITOV(ip);

	rw_destroy(&ip->i_rwlock);
	mutex_destroy(&ip->i_tlock);
	mutex_destroy(&vp->v_lock);

	cv_destroy(&vp->v_cv);
}

/* ARGSUSED */
void
ext2fs_inode_cache_reclaim(void *cdrarg)
{
}

union ihead *ihead;
kmutex_t *ih_lock;
static int ino_hashlen = 4;
int inohsz;

static int ihinitted = 0;

/*
 * Initialize hash links for inodes
 * and build inode free list.
 */
void
ihinit(void)
{
	int	i;
	union ihead *ih = ihead;

	inohsz = 1;
	ihead = kmem_zalloc(inohsz * sizeof (union ihead), KM_SLEEP);
	ih_lock = kmem_zalloc(inohsz * sizeof (kmutex_t), KM_SLEEP);

	for (i = 0, ih = ihead; i < inohsz; i++,  ih++) {
		ih->ih_head[0] = ih;
		ih->ih_head[1] = ih;
		mutex_init(&ih_lock[i], NULL, MUTEX_DEFAULT, NULL);
	}

	ext2fs_inode_cache = kmem_cache_create("ext2fs_inode_cache",
	    sizeof (struct inode), 0, ext2fs_inode_cache_constructor,
	    ext2fs_inode_cache_destructor, ext2fs_inode_cache_reclaim,
	    NULL, NULL, 0);

	ihinitted = 1;
}

static void
ihfini(void)
{
	int	i;
	union ihead *ih = ihead;
	struct inode *ip, *tip;

	if (!ihinitted)
		return;
	for (i = 0, ih = ihead; i < inohsz; i++,  ih++) {
		ip = ih->ih_chain[0];
		while (ip != (struct inode *)ih) {
			remque(ip);
			kmem_cache_free(ext2fs_inode_cache, ip);
			ip = ih->ih_chain[0];
		}
		mutex_destroy(&ih_lock[i]);
	}

	kmem_cache_destroy(ext2fs_inode_cache);
	kmem_free(ihead, inohsz * sizeof (union ihead));
	kmem_free(ih_lock, inohsz * sizeof (kmutex_t));
}

static int
ext2fs_init(struct vfssw *vswp, int fstype)
{
	vswp->vsw_vfsops = &ext2fs_vfsops;
	ext2fstype = fstype;

	ihinit();

	return (0);
}

static int
ext2fs_mount(struct vfs *vfsp, struct vnode *mvp, struct mounta *uap,
    struct cred *cr)
{
	int	error;
	struct pathname special;
	struct vnode *bvp;
	struct vnode *devvp;
	dev_t	xdev;
	struct ext2fs *fsp;
	int	needclose = 0;
	int     partno=0;
        int     pseudodev;
	int     minor;
	char    *fspec=NULL;
	
	if (!suser(cr))
		return (EPERM);
	if (mvp->v_type != VDIR)
		return (ENOTDIR);

        /* duplicate pathname because code is going to alter it */
	fspec = kmem_zalloc(strlen(uap->spec)+1, KM_SLEEP);
	strcpy(fspec, uap->spec);
	
	/* see if request wants to mount a logical partition within an extended partition */
	/* fspec has logical partition stripped off end if it is there */
	partno=ext2fs_whichpart(fspec);

        /* uap->spec is in UIO_USERSPACE, fspec is in UIO_SYSSPACE */
	error = pn_get(fspec, UIO_SYSSPACE /* UIO_USERSPACE */, &special);
	if (error)
	{
		if (fspec) 
			kmem_free(fspec, strlen(uap->spec)+1);	
		return (error);
	}
	error = lookupname(special.pn_path, UIO_SYSSPACE, FOLLOW, NULLVPP, &bvp);
	pn_free(&special);
	if (fspec) 
		kmem_free(fspec, strlen(uap->spec)+1);	

	if (error)
		return (error);
	if (bvp->v_type != VBLK) {
		VN_RELE(bvp);
		return (ENOTBLK);
	}
	xdev = bvp->v_rdev;
	VN_RELE(bvp);
	if (getmajor(xdev) >= devcnt)
		return (ENXIO);

	/*
	 * Ensure that this device (or logical drive) isn't already mounted,
	 * unless this is a REMOUNT request
	 */
	if (partno>0) {
		mutex_enter(&ext2fs_lock);
		for (fsp = ext2fs_mounttab; fsp; fsp = fsp->e_next)
			if (fsp->e_xdev == xdev &&
			    fsp->partno == partno) {
				mutex_exit(&ext2fs_lock);
				if (uap->flags & MS_REMOUNT) {
					return (0);
				} else {
					return (EBUSY);
				}
			}
		/*
		 * Assign a unique device number for the vfs
		 * The old way (getudev() + a constantly incrementing
		 * major number) was wrong because it changes vfs_dev
		 * across mounts and reboots, which breaks nfs file handles.
		 * UFS just uses the real dev_t. We can't do that because
		 * of the way ext2fs opens fdisk partitons (the :1 and :2
		 * partitions are on the same dev_t). Though that _might_
		 * actually be ok, since the file handle contains an
		 * absolute block number, it's probably better to make them
		 * different. So I think we should retain the original
		 * dev_t, but come up with a different minor number based
		 * on the logical drive that will _always_ come up the same.
		 * For now, we steal the upper 6 bits.
		 */
		minor = ((partno << 12) | getminor(xdev)) & MAXMIN32;
		pseudodev = makedevice(getmajor(xdev), minor);
		if (vfs_devismounted(pseudodev)) {
			mutex_exit(&ext2fs_lock);
			if (uap->flags & MS_REMOUNT) {
				return (0);
			} else {
				return (EBUSY);
			}
		}
		mutex_exit(&ext2fs_lock);
	} else {
		if (vfs_devismounted(xdev))
			if (uap->flags & MS_REMOUNT) {
				return (0);
			} else {
				return (EBUSY);
			}
		pseudodev = xdev;
	}


	if (uap->flags & MS_RDONLY) {
		vfsp->vfs_flag |= VFS_RDONLY;
#ifdef VSW_HASPROTO
		vfs_setmntopt(&vfsp->vfs_mntopts, MNTOPT_RO, NULL, 0);
#endif
	}

	/*
	 * Mount the filesystem
	 */
	devvp = makespecvp(xdev, VBLK);
	error = VOP_OPEN(&devvp, FREAD, cr);
	if (error)
		return (error);
	if (IS_SWAPVP(devvp)) {
		VN_RELE(devvp);
		return (EBUSY);
	}

	fsp = kmem_zalloc(sizeof (struct ext2fs), KM_SLEEP);
	fsp->e_vfs = vfsp;
	fsp->e_devvp = devvp;
	fsp->e_xdev = xdev;
	fsp->e_next = NULL;
	fsp->e_flags = 0;
        fsp->start_sector=0;
	fsp->partno=partno;
	
	if (partno>0)
	{
	  fsp->start_sector = ext2fs_get_start_sector(fsp, partno);
	}
	
	vfsp->vfs_dev = pseudodev; /* was xdev */
	vfsp->vfs_fstype = ext2fstype;
	vfs_make_fsid(&vfsp->vfs_fsid, pseudodev /* was xdev */, ext2fstype);
	vfsp->vfs_data = (caddr_t)fsp;
	vfsp->vfs_bcount = 0;

	vfsp->vfs_bsize = EXT2_MAX_BLOCK_SIZE; /* XXX - read from superblock */
	mutex_init(&fsp->e_lock, NULL, MUTEX_DEFAULT, NULL);

	/* go read and validate the super block */
	/* also reads the root inode. */
	error = ext2fs_read_super(fsp);
	if (error) {
		mutex_destroy(&fsp->e_lock);
		(void) VOP_CLOSE(devvp, FREAD, 1, (offset_t)0, cr);
		VN_RELE(devvp);
		kmem_free(fsp, sizeof (struct ext2fs));
		return (error);
	}


	mutex_enter(&ext2fs_lock);
	fsp->e_next = ext2fs_mounttab;
	ext2fs_mounttab = fsp;
	mutex_exit(&ext2fs_lock);
	return (0);
}

static int
ext2fs_unmount(struct vfs *vfsp,
#ifdef MS_FORCE
    int flag,
#endif
    struct cred *cr)
{
	struct ext2fs *fsp, *fsp1;
	union ihead *ih;
	struct inode *ip, *inext;
	struct vnode *vp;
	int	i;
	dev_t	dev = vfsp->vfs_dev;

	if (!suser(cr))
		return (EPERM);
#ifdef MS_FORCE
	if (flag & MS_FORCE)
		return (ENOTSUP);
#endif
	fsp = VFSTOEXT2FS(vfsp);
	mutex_enter(&ext2fs_lock);
	/* don't allow unmount if there are referenced inodes */
	/* only the root inode should still be referenced and in the cache */
	for (i = 0, ih = ihead; i < inohsz; i++,  ih++) {
		mutex_enter(&ih_lock[i]);
		for (inext = NULL, ip = ih->ih_chain[0];
		    ip != (struct inode *)ih; ip = inext) {
			inext = ip->i_forw;
			if (ip->i_fs != fsp)
				continue;
			vp = ITOV(ip);
			/* skip root vnode */
			if (vp == fsp->e_root)
				continue;
			mutex_exit(&ih_lock[i]);
			mutex_exit(&ext2fs_lock);
			return (EBUSY);
		}
		mutex_exit(&ih_lock[i]);
	}
	if (fsp == ext2fs_mounttab)
		ext2fs_mounttab = fsp->e_next;
	else {
		for (fsp1 = ext2fs_mounttab; fsp1 != NULL;
		    fsp1 = fsp1->e_next) {
			if (fsp1->e_next == fsp)
				fsp1->e_next = fsp1->e_next;
		}
	}
	fsp->e_next = NULL;
	/* clear the inode cache for this fs */
	for (i = 0, ih = ihead; i < inohsz; i++,  ih++) {
		mutex_enter(&ih_lock[i]);
		for (inext = NULL, ip = ih->ih_chain[0];
		    ip != (struct inode *)ih; ip = inext) {
			inext = ip->i_forw;
			if (ip->i_fs != fsp)
				continue;
			vp = ITOV(ip);
			VN_HOLD(vp);
			remque(ip);
			vp->v_type = VBAD;
			VN_RELE(vp);
			kmem_cache_free(ext2fs_inode_cache, ip);
		}
		mutex_exit(&ih_lock[i]);
	}
	mutex_exit(&ext2fs_lock);
	ext2fs_free_super(fsp);
	(void) VOP_CLOSE(fsp->e_devvp, FREAD, 1, (offset_t)0, cr);
	bflush(dev);
	(void) bfinval(dev, 1);
	VN_RELE(fsp->e_devvp);
	mutex_destroy(&fsp->e_lock);
	kmem_free(fsp, sizeof (struct ext2fs));
	return (0);
}

static int
ext2fs_root(struct vfs *vfsp, struct vnode **vpp)
{
	struct vnode *vp;
	struct ext2fs *fsp;

	if (!vfsp)
		return (EIO);
	fsp = VFSTOEXT2FS(vfsp);
	if (!fsp || !fsp->e_root)
		return (EIO);
	vp = fsp->e_root;
	VN_HOLD(vp);
	*vpp = vp;
	return (0);
}

static int
ext2fs_statvfs(struct vfs *vfsp, struct statvfs64 *sp)
{
	struct ext2fs *fsp;
	int	error;
	dev32_t	d32;
	size_t	overhead, overhead_per_group;

	fsp = VFSTOEXT2FS(vfsp);

	if (fsp == NULL)
		return (EINVAL);

	/* compute the overhead (FS structures */
	overhead_per_group = 1; /* super block */
	overhead_per_group += fsp->e_sbi.s_db_per_group; /* descriptors */
	overhead_per_group += 1;  /* block bitmap */
	overhead_per_group += 1; /* inode bitmap */
	overhead_per_group += fsp->e_sbi.s_itb_per_group;  /* inode table */

	overhead = fsp->e_sb.s_first_data_block +
	    fsp->e_sbi.s_groups_count * overhead_per_group;

	/* get information */

	bzero(sp, sizeof (*sp));
	sp->f_bsize = sp->f_frsize = EXT2_BLOCK_SIZE(fsp);
	sp->f_blocks = fsp->e_sb.s_blocks_count - overhead;
	sp->f_bfree = fsp->e_sb.s_free_blocks_count;
	if (sp->f_bfree < fsp->e_sb.s_r_blocks_count)
		sp->f_bavail = 0;
	else
		sp->f_bavail = sp->f_bfree - fsp->e_sb.s_r_blocks_count;
	sp->f_files = fsp->e_sb.s_inodes_count;
	sp->f_favail = sp->f_ffree = fsp->e_sb.s_free_inodes_count;
	(void) cmpldev(&d32, vfsp->vfs_dev);
	sp->f_fsid = d32;
	(void) strcpy(sp->f_basetype, vfssw[vfsp->vfs_fstype].vsw_name);
	sp->f_flag = vf_to_stf(vfsp->vfs_flag);
	sp->f_namemax = EXT2_NAME_LEN;

	return (0);
}

static void
ext2fs_sync_one(struct ext2fs *fsp, short flag, struct cred *cr)
{
	union ihead *ih;
	int	i;
	struct inode *ip, *inext;
	struct vnode *vp;

	if (!(flag & SYNC_CLOSE))
		return;
	bflush(fsp->e_vfs->vfs_dev);
}

static int
ext2fs_sync(struct vfs *vfsp, short flag, struct cred *cr)
{
	struct ext2fs *fsp;

	mutex_enter(&ext2fs_lock);
	if (vfsp) {
		fsp = VFSTOEXT2FS(vfsp);
		ext2fs_sync_one(fsp, flag, cr);
	} else {
		for (fsp = ext2fs_mounttab; fsp != NULL;
		    fsp = fsp->e_next) {
			ext2fs_sync_one(fsp, flag, cr);
		}
	}
	mutex_exit(&ext2fs_lock);
	return (0);
}

static int
ext2fs_vget(struct vfs *vfsp, struct vnode **vpp, struct fid *fidp)
{
	struct ext2fs *fsp = VFSTOEXT2FS(vfsp);
	struct inode	*ip;
	struct efid	*efid;
	int	error;

	if (fsp == NULL)
		goto out;

	efid = (struct efid *)fidp;
	mutex_enter(&fsp->e_lock);
	error = ext2fs_iget(vfsp, efid->efid_ino, &ip, CRED());
	mutex_exit(&fsp->e_lock);
	if (error)
		goto out;
	if ((ip->i_version != efid->efid_version) || (ip->i_mode == 0)) {
		VN_RELE(ITOV(ip));
		*vpp = NULL;
		return (0);
	}

	*vpp = ITOV(ip);
	return (0);
out:
	*vpp = NULL;
	return (0);
}
